// Firmware: "Wrfelspiel Meiern"
// Kurzbezeichnung: WM100
// ELV 09941
// Compiler: IAR Embended Workbench 4.3
// kostenlose 4K-Version erhhltlich unter
// http://supp.iar.com/Download/SW/?item=EWAVR-KS4
// (Anmeldung erforderlich)
//
// letzte nderung:  22.3.2010
// Controller: ATmega8 (Dip-Gehuse)
// Clockfrequnez = 1 MHz intern
// Author: ELV
// Version: 1.0
// Fuse-Bits:
// CKSEL=0001 SUT=10   (1MHz intern)



#include <ioavr.h>
#include <ina90.h>
#include <stdbool.h>
#include <assert.h>
#include <stdlib.h>

typedef unsigned char	uint8_t;
typedef unsigned int	uint16_t;

/* Diese Konstanten legen die Konfigurationsparameter fest. */
#define CONFIG_AUTO_POWER_OFF_DURATION_S			60
#define CONFIG_TA1_2_SWITCH_MODE_DURATION_S		 	15
#define CONFIG_TA1_MANUAL_POWER_OFF_DURATION_S		        4
#define CONFIG_TA2_DICE_DELAY_DURATION_S			5
#define CONFIG_TA2_SIMPLE_DICE_DELAY_DURATION_S		        1
#define CONFIG_TA_DEBOUNCE_DURATION_MS				30

/* Diese Makros berechnen, wie viele Timer-Ticks in einer Zeitangabe abgelaufen sind. */
#define F_CPU					         		1000000UL
#define TIMER_TICKS_PER_MS(ms)						(((F_CPU / 1000UL) * (ms)) / 64UL / 256UL)
#define TIMER_TICKS_PER_S(s)						TIMER_TICKS_PER_MS(1000UL * (s))

/* Diese Enumeration definiert die mglichen Kombinationen der zwei Taster TA1 und TA2. */
typedef enum keys
{
	NO_KEYS		= 0,
	KEY_TA1		= 1 << 0,
	KEY_TA2		= 1 << 1,
	KEYS_TA1_2	= KEY_TA1 | KEY_TA2
} keys_t;

/* Diese Struktur enthlt die Pin-Masken von Port C und D zum Schalten der LEDs,
   um mit den insgesamt 14 LEDs die Augenzahl beider Wrfel darzustellen. */
typedef struct dice_pips_to_leds
{
	uint8_t		portc;
	uint8_t		portd;
} dice_pips_to_leds_t;

/* Diese Struktur definiert einen Teil einer Melodie anhand der Anzahl Wiederholungen, Lnge und Frequenz eines Beeps.
   Mehrere Instanzen dieser Struktur bilden eine Melodie, welche mit dem Sound Transducer PZ1 abgespielt wird. */
typedef struct melody_beep
{
	uint8_t		count;
	uint8_t		frequency_ocr;
	uint16_t	duration_ms;
} melody_beep_t;

/* Dieses Makro wird zur Initialisierung der nachfolgenden LED-Tabelle verwendet. */
#define DICE_PIPS_TO_LEDS(c, d)						{ (uint8_t)(c), (uint8_t)(d) }

/* Diese Tabelle enthlt die Pin-Masken fr die Ports C und D,
   um die Ziffern 0-6 mit den LEDs als Augenzahl des linken oder rechten Wrfels anzuzeigen. */
static const dice_pips_to_leds_t s_dice_pips_to_leds[2][7] =
{
	/* rechte LED */
	{
		DICE_PIPS_TO_LEDS(0, 0),
		DICE_PIPS_TO_LEDS((1 << PORTC2), 0),
		DICE_PIPS_TO_LEDS((1 << PORTC1), (1 << PORTD0)),
		DICE_PIPS_TO_LEDS((1 << PORTC2)|(1 << PORTC1), (1 << PORTD0)),
		DICE_PIPS_TO_LEDS((1 << PORTC0)|(1 << PORTC5)|(1 << PORTC1), (1 << PORTD0)),
		DICE_PIPS_TO_LEDS((1 << PORTC0)|(1 << PORTC5)|(1 << PORTC1)|(1 << PORTC2), (1 << PORTD0)),
		DICE_PIPS_TO_LEDS((1 << PORTC0)|(1 << PORTC5)|(1 << PORTC1)|(1 << PORTC3)|(1 << PORTC4), (1 << PORTD0))
	},
	/* linke LED */
	{
		DICE_PIPS_TO_LEDS(0, 0),
		DICE_PIPS_TO_LEDS(0, (1 << PORTD4)),
		DICE_PIPS_TO_LEDS(0, (1 << PORTD1)|(1 << PORTD7)),
		DICE_PIPS_TO_LEDS(0, (1 << PORTD1)|(1 << PORTD4)|(1 << PORTD7)),
		DICE_PIPS_TO_LEDS(0, (1 << PORTD1)|(1 << PORTD3)|(1 << PORTD5)|(1 << PORTD7)),
		DICE_PIPS_TO_LEDS(0, (1 << PORTD1)|(1 << PORTD3)|(1 << PORTD5)|(1 << PORTD7)|(1 << PORTD4)),
		DICE_PIPS_TO_LEDS(0, (1 << PORTD1)|(1 << PORTD3)|(1 << PORTD5)|(1 << PORTD7)|(1 << PORTD2)|(1 << PORTD6))
	}
};

/* Definiert die Melodie, welche vorm Zeigen der Augenzahl abgespielt wird. */
static const melody_beep_t before_show_dices_pips_melody[] =
{
	{  1, 150,  300 },
	{  1, 120,  100 },
	{  1, 110,  100 },
	{  1, 100,  100 },
	{  1,   0,  500 }
};

/* Definiert die Melodie, welche beim Zeigen der Augenzahl abgespielt wird. */
static const melody_beep_t show_dices_pips_melody[] =
{
	{  1,   0,  300 },
	{  1, 100,  100 },
	{  1, 110,  100 },
	{  1, 120,  100 },
	{  1, 150,  300 },
	{  1,   0, 1000 }
};

/* Definiert die Melodie, welche beim Ausblenden der Augenzahl abgespielt wird. */
static const melody_beep_t hide_dices_pips_melody[] =
{
	{  1, 100,  100 },
	{  1, 125,  100 },
	{  1, 150,  100 }
};

/* Definiert die Melodie, welche beim erneuten Wrfeln abgespielt wird. */
static const melody_beep_t re_dice_melody[] =
{
	{  8, 250,   30 }
};

/* Definiert die Melodie, welche vorm manuellen Abschalten abgespielt wird. */
static const melody_beep_t manual_power_off_melody[] =
{
	{  3,  10,   50 }
};

/* Definiert die Melodie, welche vorm automatischen Abschalten abgespielt wird. */
static const melody_beep_t auto_power_off_melody[] =
{
	{  2,  10,   50 }
};

static bool							s_simple_mode;
static unsigned int					s_next_random_seed;
static __eeprom __no_init uint8_t	ee_initialization_magic;
static __eeprom __no_init uint8_t	ee_simple_mode;
static __eeprom __no_init uint8_t	ee_next_random_seed[2];

static void
save_next_random_seed(void)
{
	__EEPUT(&ee_next_random_seed[0], (uint8_t)((s_next_random_seed >> 0) & 0xFF));
	__EEPUT(&ee_next_random_seed[1], (uint8_t)((s_next_random_seed >> 8) & 0xFF));
}

static void
init(void)
{
	const uint8_t eeprom_init_magic = 0x9D;
	uint8_t tmp;

	PORTB	= (uint8_t)
			((1 << PORTB2)		/* Pull-Up Widerstand fr Taster TA2 aktivieren. */
			|(1 << PORTB6));	/* Pull-Up Widerstand fr Taster TA1 aktivieren. */
	DDRB	= (uint8_t)
			((0 << DDB2)		/* Taster TA2 als Eingang. */
			|(1 << DDB3)		/* Sound Transducer PZ1 als Ausgang. */
			|(0 << DDB6)		/* Taster TA1 als Eingang. */
			|(1 << DDB7));		/* Ausgang zur Steuerung von Transistor T2.*/

	/* Alle LEDs an Port C und D als Ausgang. */
	DDRC	= (uint8_t)
			((1 << DDC0)
			|(1 << DDC1)
			|(1 << DDC2)
			|(1 << DDC3)
			|(1 << DDC4)
			|(1 << DDC5));
	DDRD	= (uint8_t)
			((1 << DDD0)
			|(1 << DDD1)
			|(1 << DDD2)
			|(1 << DDD3)
			|(1 << DDD4)
			|(1 << DDD5)
			|(1 << DDD6)
			|(1 << DDD7));
	/* Alle LEDs abschalten. */
	PORTC	= DDRC;
	PORTD	= DDRD;

	/* Durch den Taster TA2 wird das Gert eingeschaltet.
	   Damit sich das Gert nicht beim Loslassen des Tasters wieder abschaltet,
	   den Transistor T2 whrend der gesamten Programmausfhrung leitend machen. */
	if ((PINB & (1 << PORTB2)) == 0)
	{
		PORTB |= (1 << PORTB7);
	}

	/* Den Timer 0 so einrichten, dass er ca. jede 16 ms den Timer-Overflow-Interrupt auslst.
	   Dies wird zum Aufwecken genutzt, um den Mikrocontroller die restliche Zeit schlafen legen zu knnen. */
	TCCR0	= (uint8_t)
			((0 << CS02)|(1 << CS01)|(1 << CS00));	/* Prescaler auf 64. */
	TIMSK	= (uint8_t)
			((1 << TOIE0));							/* Timer-Overflow-Interrupt aktivieren. */
	


	/* Im EEPROM wird der zuletzt eingestellte Spielmodus und
	   der nchste Startwert fr den Pseudo-Zufallszahlengenerator abgespeichert.
	   Dafr muss beim ersten Einschalten das EEPROM einmalig initialisiert werden.
	   Ob das EEPROM bereits initialisiert wurde, wird ber eine bestimmte Konstante geprft,
	   welche bei der Initialisierung auf einen festen Wert gesetzt wird. */
	__EEGET(tmp, &ee_initialization_magic);
	if (eeprom_init_magic == tmp)
	{
		/* EEPROM wurde bereits initialisiert. Den Spielmodus aus dem EEPROM bernehmen. */
		__EEGET(tmp, &ee_simple_mode);
		s_simple_mode = (bool)tmp;
		/* Der Zufallszahlengenerator bentigt jedesmal einen anderen Startwert, um verschiedene Zahlenreihen zu erzeugen.
		   Beim Abschalten wird ein zuflliger Wert ins EEPROM geschrieben, der hier als Startwert wieder eingelesen wird. */
		__EEGET(tmp, &ee_next_random_seed[1]);
		s_next_random_seed = (unsigned int)tmp;
		s_next_random_seed <<= 8;
		__EEGET(tmp, &ee_next_random_seed[0]);
		s_next_random_seed |= tmp;
	}
	else
	{
		/* EEPROM mit Spielmodus "Meiern" und Startwert initialisieren. */
		s_simple_mode = false;
		s_next_random_seed = 0xABBA;
		__EEPUT(&ee_simple_mode, false);
		save_next_random_seed();
		__EEPUT(&ee_initialization_magic, eeprom_init_magic);
	}

	/* Pseudo-Zufallszahlengenerator initialisieren. */
	srand(s_next_random_seed);
}

static void
delay_ms(uint16_t ms)
{
	/* Die angegebene Anzahl an Millisekunden abwarten. */
	while(ms-- != 0)
	{
		__delay_cycles(F_CPU / 1000UL);
	}
}

static keys_t
get_keys(void)
{
	/* Ermitteln, ob die Taster gerade gedrckt sind. */
	const bool keyTA1 = (PINB & (1 << PINB6)) == 0;
	const bool keyTA2 = (PINB & (1 << PINB2)) == 0;
	keys_t keys;

	/* Anhand der Taster die entsprechende Kombination ermitteln und zurckgeben. */
	if (keyTA1 && keyTA2)
	{
		keys = KEYS_TA1_2;
	}
	else if (keyTA1)
	{
		keys = KEY_TA1;
	}
	else if (keyTA2)
	{
		keys = KEY_TA2;
	}
	else
	{
		keys = NO_KEYS;
	}

	return keys;
}

static void
sound_transducer_on(const uint8_t ocr)
{
	/* Den Timer 2 so konfigurieren, dass der Sound Transducer PZ1
	   am Ausgang OC2 einen Ton in der bergebenen Frequenz erzeugt. */
	OCR2	= ocr;
	TCCR2	= (uint8_t)
			((0 << COM21)|(1 << COM20)					/* Toggle des Ausgangs OC2 bei Compare Match. */
			|(1 << WGM21)|(0 << WGM20)					/* Compare Match Modus aktivieren (CTC). */
			|(0 << CS22 )|(1 << CS21 )|(1 << CS20));	/* Timer Prescaler auf 32 setzen. */
}

static void
sound_transducer_off(void)
{
	/* Timer 2 abschalten, so dass der Sound Transducer verstummt. */
	TCCR2	= 0;
	OCR2	= 0;
	PORTB	&= ~(1 << PORTB3);
}

static void
show_dice_pips(
		const uint8_t lpips,
		const uint8_t rpips)
{
	/* Aus der Tabelle fr die Ziffern entnehmen, welche LEDs aktiviert werden mssen,
	   um die jeweilige Ziffer als Augenzahl eines Wrfels links bzw. rechts darzustellen. */
	const dice_pips_to_leds_t l = s_dice_pips_to_leds[0][lpips];
	const dice_pips_to_leds_t r = s_dice_pips_to_leds[1][rpips];
	assert(lpips <= 6 && rpips <= 6);
	/* Die beiden Ports C und D werden ausschlielich zum Anzeigen der LEDs verwendet.
	   Nachfolgend knnen die beiden Ports deshalb einfach auf die Bitmasken der zu aktivierenden LEDs gesetzt werden. */
	PORTC = (uint8_t)~(l.portc|r.portc);
	PORTD = (uint8_t)~(l.portd|r.portd);
}

static void
hide_dice_pips(void)
{
	show_dice_pips(0, 0);
}

static void
show_dice_pips_value(const uint8_t value)
{
	/* Sicherstellen, dass die Ziffern in der Quelle nach der Wertigkeit geordnet sind. */
	assert((value / 10 == 0) == (s_simple_mode != false));
	assert((value / 10 == 0) || value / 10 >= value % 10);
	/* Die beiden Ziffern sind in einer Dezimalzahl nach der Wertigkeit "gepackt".
	   Beide Ziffern nun entnehmen und die Hhere links und die Kleinere rechts anzeigen. */
	show_dice_pips(value / 10, value % 10);
}

static uint8_t
rand_dice_pips(void)
{
	/* Die Werte des Zufallszahlengenerators gleichmig auf den Wertebereich 0 bis 5 verteilen. */
	const unsigned int maxvalue = (unsigned int)rand();
	const unsigned int divider	= (unsigned int)RAND_MAX / 6u + 1u;
	uint8_t value = (uint8_t)(maxvalue / divider);
	/* Die Zufallszahl auf den Wertebereich 1 bis 6 bringen und zurckgeben. */
	++value;
	assert(1 <= value && value <= 6);
	return value;
}

static uint8_t
rand_dices_pips_value(void)
{
	uint8_t result;

	/* Eine zufllige Augenzahl beschaffen. */
	uint8_t r1 = rand_dice_pips();
	if (s_simple_mode)
	{
		/* Im einfachen Modus nur den linken Wrfel verwenden. */
		result = r1 * 10;
	}
	else
	{
		/* Im Spielmodus "Meiern" wird fr den rechten Wrfel
		   ebenfalls eine zufllige Augenzahl bentigt. */
		uint8_t r2 = rand_dice_pips();

		/* Die hhere Augenzahl immer im linken Wrfel angezeigen. */
		if (r2 > r1)
		{
			const uint8_t rt = r1;
			r1 = r2;
			r2 = rt;
		}

		/* Die Augenzahl beider Wrfel in eine Dezimalzahl verpacken. */
		result = r1 * 10 + r2;
	}

	return result;
}

static uint16_t
beep(
		uint8_t count,
		const uint8_t ocr,
		const uint16_t ms)
{
	uint16_t duration_ms = 0;
	while(count-- != 0)
	{
		/* Nur wenn eine Frequenz angegeben ist, soll ein Ton erzeugt werden.
		   Ansonsten wird die Wartezeit ohne akustisches Signal umgesetzt. */
		if (ocr != 0)
		{
			sound_transducer_on(ocr);
		}

		/* Ton in der angegebenen Dauer erzeugen. */
		delay_ms(ms);
		duration_ms += ms;

		/* Tonerzeugung abschalten. */
		sound_transducer_off();

		/* Soll derselbe Ton mehrmals hintereinander erzeugt werden,
		   nochmal dieselbe Wartezeit bis zum nchsten Ton abwarten. */
		if (count != 0)
		{
			delay_ms(ms);
			duration_ms += ms;
		}
	}

	/* Es wurde zwischengespeichert wie lange die Tonerzeugung gedauert hat.
	   Die Zeit wird in Millisekunden an den Aufrufer zurckgegeben. */
	return duration_ms;
}

static uint16_t
play_melody_beep(const melody_beep_t* const b)
{
	return beep(b->count, b->frequency_ocr, b->duration_ms);
}

static uint16_t
play_melody(
		const melody_beep_t* const beeps,
		const uint8_t beepsCount)
{
	uint8_t bi;
	uint16_t duration_ms = 0;

	/* Die einzelnen Beeps der Melodie nacheinander abspielen.
	   Die zum Abspielen der gesamten Melodie bentigte Zeit aufsummieren und zurckgeben. */
	for(bi = 0; bi != beepsCount; ++bi)
	{
		duration_ms += play_melody_beep(&beeps[bi]);
	}

	return duration_ms;
}

static void
play_wait_melody(
		uint16_t* const waitRestTicks,
		const melody_beep_t* const beeps,
		const uint8_t beepsCount)
{
	/* Melodie abspielen und dabei merken, wie lange dies dauerte. */
	const uint16_t delay = play_melody(beeps, beepsCount);

	/* Zeitangabe in Timer-Ticks umrechen. */
	const uint16_t delayTicks = (uint16_t)((delay * TIMER_TICKS_PER_MS(1000UL)) / 1000UL);

	/* Abspielzeit der Melodie von der bergebenen Wartezeit abziehen,
	   damit sich Wartezeiten reduzieren, wenn zwischenzeitlich eine Melodie abgespielt wird. */
	if (*waitRestTicks <= delayTicks)
	{
		*waitRestTicks = 0;
	}
	else
	{
		*waitRestTicks -= delayTicks;
	}
}

/* Die Beeps einer Melodie liegen als Array vor. Dieses Makro bildet daraus die Parameter des Arrays samt der Lngenangabe. */
#define melody_to_beep_params(m)	(m), (sizeof m / sizeof(m[0]))

static uint8_t
show_random_dices_pips(void)
{
	/* Eine zufllige Augenzahl fr beide Wrfel ermitteln und anzeigen. */
	const uint8_t value = rand_dices_pips_value();
	show_dice_pips_value(value);

	/* Mit dem Sound Transducer PZ1 das Wrfelgerusch erzeugen. */
	sound_transducer_on(250);
	delay_ms(10);
	sound_transducer_off();
	delay_ms(50);

	/* Augenzahl beider Wrfel zurckgeben. */
	return value;
}

static uint8_t
fade_out_random_dices_pips(const bool hide)
{
	const uint8_t rmax = rand_dice_pips() + (uint8_t)17;
	uint8_t r;
	uint8_t value;

	/* Nachdem der Taster zum Wrfeln losgelassen wurde, wird diese Funktion aufgerufen,
	   um die Wrfel in einer optischen und akustischen Animation langsam zum Stehen kommen zu lassen.
	   Die Wrfel so oft umfallen lassen, dass die Animation zwischen 2 bis 3 Sekunden andauert. */
	for(r = 0; r != rmax; ++r)
	{
		/* Zufllig wrfeln und Wrfel anzeigen. */
		(void)show_random_dices_pips();

		/* Dabei die Wartezeit nach jedem Wurf etwas erhhen,
		   damit der akustische Eindruck entsteht, dass die Wrfel bald gefallen sind. */
		delay_ms(r * 4u);

		/* Weil das Loslassen der Taste ein zuflliges Ereignis darstellt,
		   dies nutzen um den nchsten Startwert zufllig zu manipulieren. */
		s_next_random_seed += TCNT0;
	}

	delay_ms(50);
	if (hide)
	{
		/* Erneut eine zufllige Augenzahl erzeugen, jedoch ohne Anzeige. */
		hide_dice_pips();
		value = rand_dices_pips_value();
	}
	else
	{
		/* Nochmal das endgltige Fallen der Wrfel simulieren.
		   Die Augenzahl bleibt dabei angezeigt und wird zurckgeben. */
		value = show_random_dices_pips();
	}

	/* alea iacta est. */
	return value;
}

static void
VCC_off(
		const melody_beep_t* const beeps,
		const uint8_t beepsCount)
{
	/* Beim Ausschalten den nchsten Startwert fr den Zufallszahlengenerator erzeugen und diesen im EEPROM ablegen.
	   Dadurch wird sichergestellt, dass beim nchsten Start ein zuflliger Startwert bereitsteht. */
	s_next_random_seed ^= (unsigned int)rand();
	save_next_random_seed();

	/* Durch einen Sound auf die Abschaltung aufmerksam machen. */
	(void)play_melody(beeps, beepsCount);

	/* Alle LEDs abschalten. */
	hide_dice_pips();

	/* Den Transistor T2 sperren, welcher die Stromversorgung aufrecht erhlt. */
	PORTB &= ~(1 << PORTB7);

	/* Es kann einige Takte dauern, bis die Stromversorgung tatschlich abschaltet.
	   Daher wird eine Endlosschleife betreten und nicht mehr zum Aufrufer zurckgekehrt. */
	for(;;)
	{}
}

static void
yield(void)
{
	const uint8_t SMS_MASK = (uint8_t)((1 << SM2)|(1 << SM1)|(1 << SM2));
	const uint8_t SMS_IDLE = (uint8_t)((0 << SM2)|(0 << SM1)|(0 << SM2));

	/* Bis zum nchsten Timer-Tick schlafen. */
	MCUCR = (uint8_t)(((MCUCR & ~SMS_MASK) | SMS_IDLE));
	MCUCR |= (1 << SE);
	_SLEEP();
	MCUCR &=~(1 << SE);
}

int
main(void)
{
	keys_t keys = NO_KEYS;
	uint16_t keysDownTicks = 0;
	uint16_t inactivityTicks = 0;
	uint16_t waitRestTicks = 0;
	uint8_t dicePipsValue = 0;
	bool dicePipsHidden = false;
	bool pendingShowHideDicePips = false;

	init();

	_SEI();
	for(;;)
	{
		/* Wenn hier angekommen, dann ist ein weiterer Timer-Tick abgelaufen,
		   da dies in diesem Programm die einzige Ursache ist, wieso der Mikrocontroller aufwacht. */
		const keys_t lastKeys = keys;
		bool skipkeys = false;
		keys = get_keys();

		/* Je nachdem ob noch immer dieselben Tasten gedrckt sind,
		   die Zhler der Timer-Ticks erhhen bzw. zurcksetzen. */
		if (lastKeys == keys)
		{
			++keysDownTicks;
			++inactivityTicks;
		}
		else
		{
			keysDownTicks	= 0;
			inactivityTicks = 0;
		}

		/* Wartezeiten um einen Timer-Tick reduzieren. */
		if (waitRestTicks != 0)
		{
			--waitRestTicks;
		}

		/* Anhand der Haltedauer der Taster etwas zur Zuflligkeit des Startwertes beitragen. */
		if (keys != NO_KEYS)
		{
			++s_next_random_seed;
		}

		/* Muss noch die Aktion "Verdecken" oder "Aufdecken" abschlossen werden? */
		if (pendingShowHideDicePips)
		{
			assert(!s_simple_mode);

			/* Wird der Taster TA1 immer noch gehalten? */
			if ((keys & KEY_TA1) != 0)
			{
				if (!dicePipsHidden)
				{
					/* Statt den Taster loszulassen nun beide Taster gedrckt?
					   Es handelt sich dann um die Spezialfunktion des erneuten aber verdeckten Wrfelns. */
					if (keys == KEYS_TA1_2)
					{
						(void)play_melody(melody_to_beep_params(re_dice_melody));

						/* Aktion wird jetzt abgeschlossen. */
						pendingShowHideDicePips = false;

						/* Solange wrfeln bis beide Taster wieder losgelassen werden. */
						while(get_keys() != NO_KEYS)
						{
							(void)show_random_dices_pips();
						}

						/* Durch optische und akustische Animation die Wrfel fallen lassen. */
						dicePipsValue = fade_out_random_dices_pips(true);
						(void)play_melody(melody_to_beep_params(hide_dices_pips_melody));
						/* Die Augenzahl bleibt nach dem Wrfeln versteckt. */
						dicePipsHidden = true;

						skipkeys = true;
					}
				}
			}
			else
			{
				/* Aktion wird jetzt abgeschlossen. */
				pendingShowHideDicePips = false;

				if (dicePipsHidden)
				{
					/* Wrfel aufdecken und passenden Sound abspielen. */
					show_dice_pips_value(dicePipsValue);
					play_wait_melody(&waitRestTicks, melody_to_beep_params(show_dices_pips_melody));
					/* Verdecken/Aufdecken ist nur einmal mglich. */
					dicePipsValue = 0;
				}

				/* Logik umkehren, ob Wrfel beim nchstem Mal aufgedeckt oder verdeckt werden. */
				dicePipsHidden = !dicePipsHidden;

				skipkeys = true;
			}
		}

		/* Auf Tastendrcke reagieren, um das Spiel zu steuern. */
		if (!skipkeys)
		{
			switch(keys)
			{
				case NO_KEYS:
				{
					if (inactivityTicks >= (uint16_t)TIMER_TICKS_PER_S(CONFIG_AUTO_POWER_OFF_DURATION_S))
					{
						/* Nach lngerer Inaktivitt das Spiel automatisch abschalten. */
						VCC_off(melody_to_beep_params(auto_power_off_melody));
					}

					break;
				}
	
				case KEY_TA1:
				{
					if (keysDownTicks >= (uint16_t)TIMER_TICKS_PER_S(CONFIG_TA1_MANUAL_POWER_OFF_DURATION_S))
					{
						/* Wenn der Taster TA1 lange genug gehalten wird, das Gert abschalten. */
						VCC_off(melody_to_beep_params(manual_power_off_melody));
					}
					else if (keysDownTicks >= (uint16_t)TIMER_TICKS_PER_MS(CONFIG_TA_DEBOUNCE_DURATION_MS))
					{
						/* "Verdecken" und "Aufdecken" nur nach Wrfeln im Spielmodus "Meiern" verfgbar. */
						if (!pendingShowHideDicePips && !s_simple_mode && dicePipsValue != 0)
						{
							/* Sound abspielen, welche auf das "Verdecken" oder "Aufdecken" hinweist. */
							if (!dicePipsHidden)
							{
								play_wait_melody(&waitRestTicks, melody_to_beep_params(hide_dices_pips_melody));
								hide_dice_pips();
							}
							else
							{
								play_wait_melody(&waitRestTicks, melody_to_beep_params(before_show_dices_pips_melody));
							}
	
							/* Um die Tastenabfrage nicht zu blockieren, Aktion zwischenspeichern und erst beim Loslassen ausfhren. */
							pendingShowHideDicePips = true;
						}
					}
					else
					{}

					break;
				}
	
				case KEY_TA2:
				{
					/* Beim Drcken des Tasters TA2 wrfeln. */
					if (keysDownTicks >= (uint16_t)TIMER_TICKS_PER_MS(CONFIG_TA_DEBOUNCE_DURATION_MS))
					{
						/* Dies ist jedoch erst erlaubt, wenn die Wartezeit nach dem letzten Wrfeln abgelaufen ist. */
						if (0 == waitRestTicks)
						{
							/* Solange wrfeln bis der Taster wieder losgelassen wird. */
							while((get_keys() & KEY_TA2) != 0)
							{
								(void)show_random_dices_pips();
							}
		
							/* Durch optische und akustische Animation die Wrfel fallen lassen. */
							dicePipsValue = fade_out_random_dices_pips(false);
							/* Die Augenzahl bleibt nach dem Wrfeln angezeigt. */
							dicePipsHidden = false;
							pendingShowHideDicePips = false;
	
							/* Wartezeit bis zum nchsten Wrfeln je nach Spielmodus festlegen. */
							waitRestTicks	= (uint16_t)(s_simple_mode
											? TIMER_TICKS_PER_S(CONFIG_TA2_SIMPLE_DICE_DELAY_DURATION_S)
											: TIMER_TICKS_PER_S(CONFIG_TA2_DICE_DELAY_DURATION_S));
						}
					}

					break;
				}
	
				case KEYS_TA1_2:
				{
					/* Bleiben beide Taster lange genug gedrckt, so erfolgt ein Wechsel des Spielmodus. */
					if (keysDownTicks >= (uint16_t)TIMER_TICKS_PER_S(CONFIG_TA1_2_SWITCH_MODE_DURATION_S))
					{
						/* Spielmodus wechseln und im EEPROM abspeichern. */
						s_simple_mode = !s_simple_mode;
						__EEPUT(&ee_simple_mode, s_simple_mode);
	
						/* Optisch und akustisch auf die Anzahl Wrfel des Spielmodus hinweisen. */
						show_dice_pips(6, s_simple_mode ? (uint8_t)0 : (uint8_t)6);
						(void)beep(s_simple_mode ? (uint8_t)1 : (uint8_t)2, 200, 200);
	
						/* Anzeige leeren und letzten gespeicherten Wurf zurcksetzen. */
						hide_dice_pips();
						dicePipsValue = 0;
						dicePipsHidden = false;
						pendingShowHideDicePips = false;
	
						/* Warten bis einer der Taster losgelassen. */
						while(get_keys() == KEYS_TA1_2)
						{
							yield();
						}
					}

					break;
				}
			}
		}

		/* Nach jedem Schleifendurchlauf den Mikrocontroller bis zum nchsten Timer-Ticks schlafen legen. */
		yield();
	}
}

/* Dieser Interrupt wird nur zum Aufwecken des Mikrocontrollers verwendet. */
#pragma vector=TIMER0_OVF_vect
__interrupt static void
interrupt_handler_TIMER0_OVF_vect(void)
{}
